@@ -45,11 +45,14 @@ information will be user_dn or user_info.
 Additional classes can be found in the config module next to this one.
 """
 
+import copy
+from functools import reduce
 import ldap
+import operator
+import pprint
 import sys
 import traceback
-import pprint
-import copy
+import warnings
 
 from django.contrib.auth.models import User, Group, Permission
 import django.conf
@@ -70,10 +73,15 @@ except ImportError:
 # Support Django 1.5's custom user models
 try:
     from django.contrib.auth import get_user_model
-    get_user_username = lambda u: u.get_username()
+
+    def get_user_username(user):
+        return user.get_username()
 except ImportError:
-    get_user_model = lambda: User                                        # noqa
-    get_user_username = lambda u: u.username
+    def get_user_model():
+        return User
+
+    def get_user_username(user):
+        return user.username
 
 # Small compatibility hack
 try:
@@ -81,7 +89,7 @@ try:
 except NameError:
     basestring = str
 
-from django_auth_ldap.config import _LDAPConfig, LDAPSearch
+from django_auth_ldap.config import ConfigurationWarning, _LDAPConfig, LDAPGroupQuery, LDAPSearch
 
 
 logger = _LDAPConfig.get_logger()
@@ -158,13 +166,13 @@ class LDAPBackend(object):
     # The Django auth backend API
     #
 
-    def authenticate(self, username, password, **kwargs):
-        if len(password) == 0 and not self.settings.PERMIT_EMPTY_PASSWORD:
+    def authenticate(self, request=None, username=None, password=None, **kwargs):
+        if bool(password) or self.settings.PERMIT_EMPTY_PASSWORD:
+            ldap_user = _LDAPUser(self, username=username.strip())
+            user = ldap_user.authenticate(password)
+        else:
             logger.debug('Rejecting empty password for %s' % username)
-            return None
-
-        ldap_user = _LDAPUser(self, username=username.strip())
-        user = ldap_user.authenticate(password)
+            user = None
 
         return user
 
@@ -456,7 +464,7 @@ class _LDAPUser(object):
 
     def _load_user_attrs(self):
         if self.dn is not None:
-            search = LDAPSearch(self.dn, ldap.SCOPE_BASE)
+            search = LDAPSearch(self.dn, ldap.SCOPE_BASE, attrlist=self.settings.USER_ATTRLIST)
             results = search.execute(self.connection)
 
             if results is not None and len(results) > 0:
@@ -464,14 +472,21 @@ class _LDAPUser(object):
 
     def _load_user_dn(self):
         """
-        Populates self._user_dn with the distinguished name of our user. This
-        will either construct the DN from a template in
+        Populates self._user_dn with the distinguished name of our user.
+
+        This will either construct the DN from a template in
         AUTH_LDAP_USER_DN_TEMPLATE or connect to the server and search for it.
+        If we have to search, we'll cache the DN.
+
         """
         if self._using_simple_bind_mode():
-            self._construct_simple_user_dn()
+            self._user_dn = self._construct_simple_user_dn()
         else:
-            self._search_for_user_dn()
+            cache_key = 'django_auth_ldap.user_dn.{}'.format(self._username)
+            self._user_dn = cache_get_or_set(
+                cache, cache_key, self._search_for_user_dn,
+                self.settings.GROUP_CACHE_TIMEOUT
+            )
 
     def _using_simple_bind_mode(self):
         return (self.settings.USER_DN_TEMPLATE is not None)
@@ -480,7 +495,9 @@ class _LDAPUser(object):
         template = self.settings.USER_DN_TEMPLATE
         username = ldap.dn.escape_dn_chars(self._username)
 
-        self._user_dn = template % {'user': username}
+        user_dn = template % {'user': username}
+
+        return user_dn
 
     def _search_for_user_dn(self):
         """
@@ -492,8 +509,12 @@ class _LDAPUser(object):
             raise ImproperlyConfigured('AUTH_LDAP_USER_SEARCH must be an LDAPSearch instance.')
 
         results = search.execute(self.connection, {'user': self._username})
-        if results is not None and len(results) == 1:
-            (self._user_dn, self._user_attrs) = next(iter(results))
+        if (results is not None) and (len(results) == 1):
+            (user_dn, self._user_attrs) = next(iter(results))
+        else:
+            user_dn = None
+
+        return user_dn
 
     def _check_requirements(self):
         """
@@ -511,9 +532,11 @@ class _LDAPUser(object):
         required_group_dn = self.settings.REQUIRE_GROUP
 
         if required_group_dn is not None:
-            is_member = self._get_groups().is_member_of(required_group_dn)
-            if not is_member:
-                raise self.AuthenticationFailed("user is not a member of AUTH_LDAP_REQUIRE_GROUP")
+            if not isinstance(required_group_dn, LDAPGroupQuery):
+                required_group_dn = LDAPGroupQuery(required_group_dn)
+            result = required_group_dn.resolve(self)
+            if not result:
+                raise self.AuthenticationFailed("user does not satisfy AUTH_LDAP_REQUIRE_GROUP")
 
         return True
 
@@ -527,7 +550,7 @@ class _LDAPUser(object):
         if denied_group_dn is not None:
             is_member = self._get_groups().is_member_of(denied_group_dn)
             if is_member:
-                raise self.AuthenticationFailed("user is a member of AUTH_LDAP_DENY_GROUP")
+                raise self.AuthenticationFailed("user does not satisfy AUTH_LDAP_DENY_GROUP")
 
         return True
 
@@ -561,7 +584,8 @@ class _LDAPUser(object):
             self._populate_user()
             save_user = True
 
-        if self.settings.MIRROR_GROUPS:
+        if bool(self.settings.MIRROR_GROUPS) or bool(self.settings.MIRROR_GROUPS_EXCEPT):
+            self._normalize_mirror_settings()
             self._mirror_groups()
 
         # Give the client a chance to finish populating the user just before
@@ -596,9 +620,12 @@ class _LDAPUser(object):
 
     def _populate_user_from_group_memberships(self):
         for field, group_dns in self.settings.USER_FLAGS_BY_GROUP.items():
-            if isinstance(group_dns, basestring):
-                group_dns = [group_dns]
-            value = any(self._get_groups().is_member_of(dn) for dn in group_dns)
+            try:
+                query = self._normalize_group_dns(group_dns)
+            except ValueError as e:
+                raise ImproperlyConfigured("{}: {}", self.settings._name('USER_FLAGS_BY_GROUP'), e)
+
+            value = query.resolve(self)
             setattr(self._user, field, value)
 
     def _should_populate_profile(self):
@@ -653,14 +680,86 @@ class _LDAPUser(object):
         save_profile = False
 
         for field, group_dns in self.settings.PROFILE_FLAGS_BY_GROUP.items():
-            if isinstance(group_dns, basestring):
-                group_dns = [group_dns]
-            value = any(self._get_groups().is_member_of(dn) for dn in group_dns)
+            try:
+                query = self._normalize_group_dns(group_dns)
+            except ValueError as e:
+                raise ImproperlyConfigured("{}: {}", self.settings._name('PROFILE_FLAGS_BY_GROUP'), e)
+
+            value = query.resolve(self)
             setattr(profile, field, value)
             save_profile = True
 
         return save_profile
 
+    def _normalize_group_dns(self, group_dns):
+        """
+        Converts one or more group DNs to an LDAPGroupQuery.
+
+        group_dns may be a string, a non-empty list or tuple of strings, or an
+        LDAPGroupQuery. The result will be an LDAPGroupQuery. A list or tuple
+        will be joined with the | operator.
+
+        """
+        if isinstance(group_dns, LDAPGroupQuery):
+            query = group_dns
+        elif isinstance(group_dns, basestring):
+            query = LDAPGroupQuery(group_dns)
+        elif isinstance(group_dns, (list, tuple)) and len(group_dns) > 0:
+            query = reduce(operator.or_, map(LDAPGroupQuery, group_dns))
+        else:
+            raise ValueError(group_dns)
+
+        return query
+
+    def _normalize_mirror_settings(self):
+        """
+        Validates the group mirroring settings and converts them as necessary.
+        """
+        def malformed_mirror_groups_except():
+            return ImproperlyConfigured(
+                "{} must be a collection of group names".format(
+                    self.settings._name('MIRROR_GROUPS_EXCEPT')
+                )
+            )
+
+        def malformed_mirror_groups():
+            return ImproperlyConfigured(
+                "{} must be True or a collection of group names".format(
+                    self.settings._name('MIRROR_GROUPS')
+                )
+            )
+
+        mge = self.settings.MIRROR_GROUPS_EXCEPT
+        mg = self.settings.MIRROR_GROUPS
+
+        if mge is not None:
+            if isinstance(mge, (set, frozenset)):
+                pass
+            elif isinstance(mge, (list, tuple)):
+                mge = self.settings.MIRROR_GROUPS_EXCEPT = frozenset(mge)
+            else:
+                raise malformed_mirror_groups_except()
+
+            if not all(isinstance(value, basestring) for value in mge):
+                raise malformed_mirror_groups_except()
+            elif bool(mg):
+                warnings.warn(ConfigurationWarning("Ignoring {} in favor of {}".format(
+                    self.settings._name("MIRROR_GROUPS"),
+                    self.settings._name("MIRROR_GROUPS_EXCEPT")
+                )))
+                mg = self.settings.MIRROR_GROUPS = None
+
+        if mg is not None:
+            if isinstance(mg, (bool, set, frozenset)):
+                pass
+            elif isinstance(mg, (list, tuple)):
+                mg = self.settings.MIRROR_GROUPS = frozenset(mg)
+            else:
+                raise malformed_mirror_groups()
+
+            if isinstance(mg, (set, frozenset)) and (not all(isinstance(value, basestring) for value in mg)):
+                raise malformed_mirror_groups()
+
     def _mirror_groups(self):
         """
         Mirrors the user's LDAP groups in the Django database and updates the
@@ -669,6 +768,24 @@ class _LDAPUser(object):
         target_group_names = frozenset(self._get_groups().get_group_names())
         current_group_names = frozenset(self._user.groups.values_list('name', flat=True).iterator())
 
+        # These were normalized to sets above.
+        MIRROR_GROUPS_EXCEPT = self.settings.MIRROR_GROUPS_EXCEPT
+        MIRROR_GROUPS = self.settings.MIRROR_GROUPS
+
+        # If the settings are white- or black-listing groups, we'll update
+        # target_group_names such that we won't modify the membership of groups
+        # beyond our purview.
+        if isinstance(MIRROR_GROUPS_EXCEPT, (set, frozenset)):
+            target_group_names = (
+                (target_group_names - MIRROR_GROUPS_EXCEPT) |
+                (current_group_names & MIRROR_GROUPS_EXCEPT)
+            )
+        elif isinstance(MIRROR_GROUPS, (set, frozenset)):
+            target_group_names = (
+                (target_group_names & MIRROR_GROUPS) |
+                (current_group_names - MIRROR_GROUPS)
+            )
+
         if target_group_names != current_group_names:
             existing_groups = list(Group.objects.filter(name__in=target_group_names).iterator())
             existing_group_names = frozenset(group.name for group in existing_groups)
@@ -769,8 +886,10 @@ class _LDAPUserGroups(object):
 
     def _init_group_settings(self):
         """
-        Loads the settings we need to deal with groups. Raises
-        ImproperlyConfigured if anything's not right.
+        Loads the settings we need to deal with groups.
+
+        Raises ImproperlyConfigured if anything's not right.
+
         """
         self._group_type = self.settings.GROUP_TYPE
         if self._group_type is None:
@@ -870,6 +989,8 @@ class LDAPSettings(object):
     instance will contain all of our settings as attributes, with default values
     if they are not specified by the configuration.
     """
+    _prefix = 'AUTH_LDAP_'
+
     defaults = {
         'ALWAYS_UPDATE_USER': True,
         'AUTHORIZE_ALL_USERS': False,
@@ -883,13 +1004,15 @@ class LDAPSettings(object):
         'GROUP_CACHE_TIMEOUT': None,
         'GROUP_SEARCH': None,
         'GROUP_TYPE': None,
-        'MIRROR_GROUPS': False,
+        'MIRROR_GROUPS': None,
+        'MIRROR_GROUPS_EXCEPT': None,
         'PERMIT_EMPTY_PASSWORD': False,
         'PROFILE_ATTR_MAP': {},
         'PROFILE_FLAGS_BY_GROUP': {},
         'REQUIRE_GROUP': None,
         'SERVER_URI': 'ldap://localhost',
         'START_TLS': False,
+        'USER_ATTRLIST': None,
         'USER_ATTR_MAP': {},
         'USER_DN_TEMPLATE': None,
         'USER_FLAGS_BY_GROUP': {},
@@ -901,8 +1024,25 @@ class LDAPSettings(object):
         Loads our settings from django.conf.settings, applying defaults for any
         that are omitted.
         """
+        self._prefix = prefix
+
         defaults = dict(self.defaults, **defaults)
 
         for name, default in defaults.items():
             value = getattr(django.conf.settings, prefix + name, default)
             setattr(self, name, value)
+
+    def _name(self, suffix):
+        return (self._prefix + suffix)
+
+
+def cache_get_or_set(cache, key, default, timeout=None):
+    """
+    Backport of Django 1.9's cache.get_or_set.
+    """
+    value = cache.get(key)
+    if value is None:
+        value = default() if callable(default) else default
+        cache.set(key, value, timeout)
+
+    return value
