@@ -34,12 +34,17 @@ import ldap
 import logging
 import pprint
 
+from django.utils.tree import Node
 try:
     from django.utils.encoding import force_str
 except ImportError:  # Django < 1.5
     from django.utils.encoding import smart_str as force_str
 
 
+class ConfigurationWarning(UserWarning):
+    pass
+
+
 class _LDAPConfig(object):
     """
     A private class that loads and caches some global objects.
@@ -105,7 +110,7 @@ class LDAPSearch(object):
     documented for configuration purposes. Internal clients may use the other
     methods to refine and execute the search.
     """
-    def __init__(self, base_dn, scope, filterstr=u'(objectClass=*)'):
+    def __init__(self, base_dn, scope, filterstr=u'(objectClass=*)', attrlist=None):
         """
         These parameters are the same as the first three parameters to
         ldap.search_s.
@@ -113,8 +118,12 @@ class LDAPSearch(object):
         self.base_dn = base_dn
         self.scope = scope
         self.filterstr = filterstr
+        self.attrlist = attrlist
         self.ldap = _LDAPConfig.get_ldap()
 
+    def __repr__(self):
+        return "<{}: {}>".format(self.__class__.__name__, self.base_dn)
+
     def search_with_additional_terms(self, term_dict, escape=True):
         """
         Returns a new search object with additional search terms and-ed to the
@@ -159,7 +168,8 @@ class LDAPSearch(object):
             filterstr = self.filterstr % filterargs
             results = connection.search_s(force_str(self.base_dn),
                                           self.scope,
-                                          force_str(filterstr))
+                                          force_str(filterstr),
+                                          self.attrlist)
         except ldap.LDAPError as e:
             results = []
             logger.error(u"search_s('%s', %d, '%s') raised %s" %
@@ -182,7 +192,8 @@ class LDAPSearch(object):
         try:
             filterstr = self.filterstr % filterargs
             msgid = connection.search(force_str(self.base_dn),
-                                      self.scope, force_str(filterstr))
+                                      self.scope, force_str(filterstr),
+                                      self.attrlist)
         except ldap.LDAPError as e:
             msgid = None
             logger.error(u"search('%s', %d, '%s') raised %s" %
@@ -196,7 +207,7 @@ class LDAPSearch(object):
         """
         try:
             kind, results = connection.result(msgid)
-            if kind != ldap.RES_SEARCH_RESULT:
+            if kind not in (ldap.RES_SEARCH_ENTRY, ldap.RES_SEARCH_RESULT):
                 results = []
         except ldap.LDAPError as e:
             results = []
@@ -452,6 +463,9 @@ class MemberDNGroupType(LDAPGroupType):
 
         super(MemberDNGroupType, self).__init__(name_attr)
 
+    def __repr__(self):
+        return "<{}: {}>".format(self.__class__.__name__, self.member_attr)
+
     def user_groups(self, ldap_user, group_search):
         search = group_search.search_with_additional_terms({self.member_attr: ldap_user.dn})
         groups = search.execute(ldap_user.connection)
@@ -626,3 +640,86 @@ class NestedOrganizationalRoleGroupType(NestedMemberDNGroupType):
     """
     def __init__(self, name_attr='cn'):
         super(NestedOrganizationalRoleGroupType, self).__init__('roleOccupant', name_attr)
+
+
+class LDAPGroupQuery(Node):
+    """
+    Represents a compound query for group membership.
+
+    This can be used to construct an arbitrarily complex group membership query
+    with AND, OR, and NOT logical operators. Construct primitive queries with a
+    group DN as the only argument. These queries can then be combined with the
+    ``&``, ``|``, and ``~`` operators.
+
+    :param str group_dn: The DN of a group to test for membership.
+
+    """
+    # Connection types
+    AND = 'AND'
+    OR = 'OR'
+    default = AND
+
+    _CONNECTORS = [AND, OR]
+
+    def __init__(self, *args, **kwargs):
+        super(LDAPGroupQuery, self).__init__(children=list(args) + list(kwargs.items()))
+
+    def __and__(self, other):
+        return self._combine(other, self.AND)
+
+    def __or__(self, other):
+        return self._combine(other, self.OR)
+
+    def __invert__(self):
+        obj = type(self)()
+        obj.add(self, self.AND)
+        obj.negate()
+
+        return obj
+
+    def _combine(self, other, conn):
+        if not isinstance(other, LDAPGroupQuery):
+            raise TypeError(other)
+        if conn not in self._CONNECTORS:
+            raise ValueError(conn)
+
+        obj = type(self)()
+        obj.connector = conn
+        obj.add(self, conn)
+        obj.add(other, conn)
+
+        return obj
+
+    def resolve(self, ldap_user, groups=None):
+        if groups is None:
+            groups = ldap_user._get_groups()
+
+        result = self.aggregator(self._resolve_children(ldap_user, groups))
+        if self.negated:
+            result = not result
+
+        return result
+
+    @property
+    def aggregator(self):
+        """
+        Returns a function for aggregating a sequence of sub-results.
+        """
+        if self.connector == self.AND:
+            aggregator = all
+        elif self.connector == self.OR:
+            aggregator = any
+        else:
+            raise ValueError(self.connector)
+
+        return aggregator
+
+    def _resolve_children(self, ldap_user, groups):
+        """
+        Generates the query result for each child.
+        """
+        for child in self.children:
+            if isinstance(child, LDAPGroupQuery):
+                yield child.resolve(ldap_user, groups)
+            else:
+                yield groups.is_member_of(child)
