@@ -38,8 +38,10 @@ except ImportError:
 
 import django
 from django.conf import settings
-import django.db.models.signals
 from django.contrib.auth.models import User, Permission, Group
+from django.core.cache import cache
+from django.core.exceptions import ImproperlyConfigured
+import django.db.models.signals
 from django.test import TestCase
 
 try:
@@ -55,10 +57,12 @@ except ImportError:
 try:
     from django.test.utils import override_settings
 except ImportError:
-    override_settings = lambda *args, **kwargs: (lambda v: v)
+    def override_settings(*args, **kwargs):
+        return lambda v: v
 
 from django_auth_ldap.models import TestUser, TestProfile
 from django_auth_ldap import backend
+from django_auth_ldap.config import LDAPGroupQuery
 from django_auth_ldap.config import LDAPSearch, LDAPSearchUnion
 from django_auth_ldap.config import PosixGroupType, MemberDNGroupType, NestedMemberDNGroupType, NISGroupType
 from django_auth_ldap.config import GroupOfNamesType
@@ -81,6 +85,7 @@ class LDAPTest(TestCase):
     people = ("ou=people,o=test", {"ou": "people"})
     groups = ("ou=groups,o=test", {"ou": "groups"})
     moregroups = ("ou=moregroups,o=test", {"ou": "moregroups"})
+    mirror_groups = ("ou=mirror_groups,o=test", {"ou": "mirror_groups"})
 
     alice = ("uid=alice,ou=people,o=test", {
         "uid": ["alice"],
@@ -163,6 +168,45 @@ class LDAPTest(TestCase):
         "member": ["uid=bob,ou=people,o=test"]
     })
 
+    # groupOfNames objects for LDAPGroupQuery testing
+    alice_gon = ("cn=alice_gon,ou=query_groups,o=test", {
+        "cn": ["alice_gon"],
+        "objectClass": ["groupOfNames"],
+        "member": ["uid=alice,ou=people,o=test"]
+    })
+    mutual_gon = ("cn=mutual_gon,ou=query_groups,o=test", {
+        "cn": ["mutual_gon"],
+        "objectClass": ["groupOfNames"],
+        "member": ["uid=alice,ou=people,o=test", "uid=bob,ou=people,o=test"]
+    })
+    bob_gon = ("cn=bob_gon,ou=query_groups,o=test", {
+        "cn": ["bob_gon"],
+        "objectClass": ["groupOfNames"],
+        "member": ["uid=bob,ou=people,o=test"]
+    })
+
+    # groupOfNames objects for selective group mirroring.
+    mirror1 = ("cn=mirror1,ou=mirror_groups,o=test", {
+        "cn": ["mirror1"],
+        "objectClass": ["groupOfNames"],
+        "member": ["uid=alice,ou=people,o=test"]
+    })
+    mirror2 = ("cn=mirror2,ou=mirror_groups,o=test", {
+        "cn": ["mirror2"],
+        "objectClass": ["groupOfNames"],
+        "member": []
+    })
+    mirror3 = ("cn=mirror3,ou=mirror_groups,o=test", {
+        "cn": ["mirror3"],
+        "objectClass": ["groupOfNames"],
+        "member": ["uid=alice,ou=people,o=test"]
+    })
+    mirror4 = ("cn=mirror4,ou=mirror_groups,o=test", {
+        "cn": ["mirror4"],
+        "objectClass": ["groupOfNames"],
+        "member": []
+    })
+
     # nisGroup objects
     active_nis = ("cn=active_nis,ou=groups,o=test", {
         "cn": ["active_nis"],
@@ -200,11 +244,13 @@ class LDAPTest(TestCase):
         "member": ["cn=parent_gon,ou=groups,o=test"]
     })
 
-    directory = dict([top, people, groups, moregroups, alice, bob, dressler,
-                      nobody, active_px, staff_px, superuser_px, empty_gon,
-                      active_gon, staff_gon, superuser_gon, other_gon,
-                      active_nis, staff_nis, superuser_nis,
-                      parent_gon, nested_gon, circular_gon])
+    directory = dict([
+        top, people, groups, moregroups, mirror_groups, alice, bob, dressler,
+        nobody, active_px, staff_px, superuser_px, empty_gon, active_gon,
+        staff_gon, superuser_gon, other_gon, alice_gon, mutual_gon, bob_gon,
+        mirror1, mirror2, mirror3, mirror4, active_nis, staff_nis,
+        superuser_nis, parent_gon, nested_gon, circular_gon
+    ])
 
     @classmethod
     def configure_logger(cls):
@@ -223,13 +269,20 @@ class LDAPTest(TestCase):
         cls.configure_logger()
         cls.mockldap = mockldap.MockLdap(cls.directory)
 
-        warnings.filterwarnings('ignore', message='.*?AUTH_PROFILE_MODULE', category=DeprecationWarning, module='django_auth_ldap')
+        warnings.filterwarnings(
+            'ignore', message='.*?AUTH_PROFILE_MODULE', category=DeprecationWarning, module='django_auth_ldap'
+        )
+        warnings.filterwarnings(
+            'ignore', category=UnicodeWarning, module='mockldap.ldapobject'
+        )
 
     @classmethod
     def tearDownClass(cls):
         del cls.mockldap
 
     def setUp(self):
+        cache.clear()
+
         self.mockldap.start()
         self.ldapobj = self.mockldap['ldap://localhost']
 
@@ -303,6 +356,16 @@ class LDAPTest(TestCase):
             ['initialize', 'simple_bind_s']
         )
 
+    @override_settings(AUTH_LDAP_USER_SEARCH=LDAPSearch("ou=people,o=test", ldap.SCOPE_SUBTREE, '(uid=%(user)s)'))
+    def test_login_with_multiple_auth_backends(self):
+        auth = self.client.login(username='alice', password='password')
+        self.assertTrue(auth)
+
+    @override_settings(AUTH_LDAP_USER_SEARCH=LDAPSearch("ou=people,o=test", ldap.SCOPE_SUBTREE, '(uid=%(user)s)'))
+    def test_bad_login_with_multiple_auth_backends(self):
+        auth = self.client.login(username='invalid', password='i_do_not_exist')
+        self.assertFalse(auth)
+
     def test_simple_bind_escaped(self):
         """ Bind with a username that requires escaping. """
         self._init_settings(
@@ -427,7 +490,7 @@ class LDAPTest(TestCase):
         )
         # mockldap doesn't handle case-insensitive matching properly.
         self.ldapobj.search_s.seed('ou=people,o=test', ldap.SCOPE_SUBTREE,
-                                   '(uid=Alice)')([self.alice])
+                                   '(uid=Alice)', None)([self.alice])
         User.objects.create(username='alice')
 
         user = self.backend.authenticate(username='Alice', password='password')
@@ -493,7 +556,7 @@ class LDAPTest(TestCase):
             self.ldapobj.methods_called(with_args=True),
             [('initialize', ('ldap://localhost',), {}),
              ('simple_bind_s', ('', ''), {}),
-             ('search_s', ('ou=people,o=test', ldap.SCOPE_SUBTREE, '(uid=alice\\2a)'), {})]
+             ('search_s', ('ou=people,o=test', ldap.SCOPE_SUBTREE, '(uid=alice\\2a)', None), {})]
         )
 
     def test_search_bind_no_user(self):
@@ -616,6 +679,29 @@ class LDAPTest(TestCase):
             ['initialize', 'simple_bind_s', 'simple_bind_s', 'search_s']
         )
 
+    def test_populate_with_attrlist(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            USER_ATTR_MAP={'first_name': 'givenName', 'last_name': 'sn'},
+            USER_ATTRLIST=['*', '+'],
+        )
+
+        user = self.backend.authenticate(username='alice', password='password')
+
+        self.assertEqual(user.username, 'alice')
+        self.assertEqual(user.first_name, 'Alice')
+        self.assertEqual(user.last_name, 'Adams')
+
+        # init, bind as user, bind anonymous, lookup user attrs
+        self.assertEqual(
+            self.ldapobj.methods_called(),
+            ['initialize', 'simple_bind_s', 'simple_bind_s', 'search_s']
+        )
+        self.assertEqual(
+            self.ldapobj.methods_called(with_args=True)[3],
+            ('search_s', ('uid=alice,ou=people,o=test', ldap.SCOPE_BASE, '(objectClass=*)', ['*', '+']), {})
+        )
+
     def test_bind_as_user(self):
         self._init_settings(
             USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
@@ -729,6 +815,95 @@ class LDAPTest(TestCase):
              'initialize', 'simple_bind_s', 'simple_bind_s', 'compare_s']
         )
 
+    def test_simple_group_query(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=query_groups,o=test', ldap.SCOPE_SUBTREE, '(objectClass=groupOfNames)'),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+        )
+        alice = self.backend.authenticate(username='alice', password='password')
+        query = LDAPGroupQuery('cn=alice_gon,ou=query_groups,o=test')
+        self.assertTrue(query.resolve(alice.ldap_user))
+
+    def test_negated_group_query(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=query_groups,o=test', ldap.SCOPE_SUBTREE, '(objectClass=groupOfNames)'),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+        )
+        alice = self.backend.authenticate(username='alice', password='password')
+        query = ~LDAPGroupQuery('cn=alice_gon,ou=query_groups,o=test')
+        self.assertFalse(query.resolve(alice.ldap_user))
+
+    def test_or_group_query(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=query_groups,o=test', ldap.SCOPE_SUBTREE, '(objectClass=groupOfNames)'),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+        )
+        alice = self.backend.authenticate(username='alice', password='password')
+        bob = self.backend.authenticate(username='bob', password='password')
+
+        query = (
+            LDAPGroupQuery('cn=alice_gon,ou=query_groups,o=test') |
+            LDAPGroupQuery('cn=bob_gon,ou=query_groups,o=test')
+        )
+        self.assertTrue(query.resolve(alice.ldap_user))
+        self.assertTrue(query.resolve(bob.ldap_user))
+
+    def test_and_group_query(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=query_groups,o=test', ldap.SCOPE_SUBTREE, '(objectClass=groupOfNames)'),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+        )
+        alice = self.backend.authenticate(username='alice', password='password')
+        bob = self.backend.authenticate(username='bob', password='password')
+
+        query = (
+            LDAPGroupQuery('cn=alice_gon,ou=query_groups,o=test') &
+            LDAPGroupQuery('cn=mutual_gon,ou=query_groups,o=test')
+        )
+        self.assertTrue(query.resolve(alice.ldap_user))
+        self.assertFalse(query.resolve(bob.ldap_user))
+
+    def test_nested_group_query(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=query_groups,o=test', ldap.SCOPE_SUBTREE, '(objectClass=groupOfNames)'),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+        )
+        alice = self.backend.authenticate(username='alice', password='password')
+        bob = self.backend.authenticate(username='bob', password='password')
+
+        query = (
+            (
+                LDAPGroupQuery('cn=alice_gon,ou=query_groups,o=test') &
+                LDAPGroupQuery('cn=mutual_gon,ou=query_groups,o=test')
+            ) |
+            LDAPGroupQuery('cn=bob_gon,ou=query_groups,o=test')
+        )
+        self.assertTrue(query.resolve(alice.ldap_user))
+        self.assertTrue(query.resolve(bob.ldap_user))
+
+    def test_require_group_as_group_query(self):
+        query = (
+            LDAPGroupQuery('cn=alice_gon,ou=query_groups,o=test') &
+            LDAPGroupQuery('cn=mutual_gon,ou=query_groups,o=test')
+        )
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=query_groups,o=test', ldap.SCOPE_SUBTREE, '(objectClass=groupOfNames)'),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+            REQUIRE_GROUP=query
+        )
+
+        alice = self.backend.authenticate(username='alice', password='password')
+        bob = self.backend.authenticate(username='bob', password='password')
+
+        self.assertTrue(alice is not None)
+        self.assertTrue(bob is None)
+
     def test_group_union(self):
         self._init_settings(
             USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
@@ -792,7 +967,10 @@ class LDAPTest(TestCase):
         )
         alice = self.backend.authenticate(username='alice', password='password')
 
-        self.assertEqual(alice.ldap_user.group_dns, set((g[0].lower() for g in [self.active_gon, self.staff_gon, self.superuser_gon, self.nested_gon])))
+        self.assertEqual(
+            alice.ldap_user.group_dns,
+            set((g[0].lower() for g in [self.active_gon, self.staff_gon, self.superuser_gon, self.nested_gon]))
+        )
 
     def test_group_names(self):
         self._init_settings(
@@ -810,7 +988,7 @@ class LDAPTest(TestCase):
             GROUP_SEARCH=LDAPSearch('ou=groups,o=test', ldap.SCOPE_SUBTREE),
             GROUP_TYPE=MemberDNGroupType(member_attr='member'),
             USER_FLAGS_BY_GROUP={
-                'is_active': "cn=active_gon,ou=groups,o=test",
+                'is_active': LDAPGroupQuery("cn=active_gon,ou=groups,o=test"),
                 'is_staff': ["cn=empty_gon,ou=groups,o=test",
                              "cn=staff_gon,ou=groups,o=test"],
                 'is_superuser': "cn=superuser_gon,ou=groups,o=test"
@@ -827,6 +1005,21 @@ class LDAPTest(TestCase):
         self.assertTrue(not bob.is_staff)
         self.assertTrue(not bob.is_superuser)
 
+    def test_user_flags_misconfigured(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=groups,o=test', ldap.SCOPE_SUBTREE),
+            GROUP_TYPE=MemberDNGroupType(member_attr='member'),
+            USER_FLAGS_BY_GROUP={
+                'is_active': LDAPGroupQuery("cn=active_gon,ou=groups,o=test"),
+                'is_staff': [],
+                'is_superuser': "cn=superuser_gon,ou=groups,o=test"
+            }
+        )
+
+        with self.assertRaises(ImproperlyConfigured):
+            self.backend.authenticate(username='alice', password='password')
+
     def test_posix_membership(self):
         self._init_settings(
             USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
@@ -1096,6 +1289,105 @@ class LDAPTest(TestCase):
         )
         self.assertEqual(set(alice.groups.all()), set(Group.objects.all()))
 
+    #
+    # When selectively mirroring groups, there are eight scenarios for any
+    # given user/group pair:
+    #
+    #   (is-member-in-LDAP, not-member-in-LDAP)
+    #   x (is-member-in-Django, not-member-in-Django)
+    #   x (synced, not-synced)
+    #
+    # The four test cases below take these scenarios four at a time for each of
+    # the two settings.
+
+    def test_group_mirroring_whitelist_update(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=mirror_groups,o=test', ldap.SCOPE_SUBTREE,
+                                    '(objectClass=groupOfNames)'),
+            GROUP_TYPE=GroupOfNamesType(),
+            MIRROR_GROUPS=['mirror1', 'mirror2']
+        )
+
+        groups = {}
+        for name in ('mirror{}'.format(i) for i in range(1, 5)):
+            groups[name] = Group.objects.create(name=name)
+        alice = self.backend.populate_user('alice')
+        alice.groups = [groups['mirror2'], groups['mirror4']]
+
+        alice = self.backend.authenticate(username='alice', password='password')
+
+        self.assertEqual(
+            set(alice.groups.values_list("name", flat=True)),
+            {'mirror1', 'mirror4'}
+        )
+
+    def test_group_mirroring_whitelist_noop(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=mirror_groups,o=test', ldap.SCOPE_SUBTREE,
+                                    '(objectClass=groupOfNames)'),
+            GROUP_TYPE=GroupOfNamesType(),
+            MIRROR_GROUPS=['mirror1', 'mirror2']
+        )
+
+        groups = {}
+        for name in ('mirror{}'.format(i) for i in range(1, 5)):
+            groups[name] = Group.objects.create(name=name)
+        alice = self.backend.populate_user('alice')
+        alice.groups = [groups['mirror1'], groups['mirror3']]
+
+        alice = self.backend.authenticate(username='alice', password='password')
+
+        self.assertEqual(
+            set(alice.groups.values_list("name", flat=True)),
+            {'mirror1', 'mirror3'}
+        )
+
+    def test_group_mirroring_blacklist_update(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=mirror_groups,o=test', ldap.SCOPE_SUBTREE,
+                                    '(objectClass=groupOfNames)'),
+            GROUP_TYPE=GroupOfNamesType(),
+            MIRROR_GROUPS_EXCEPT=['mirror1', 'mirror2']
+        )
+
+        groups = {}
+        for name in ('mirror{}'.format(i) for i in range(1, 5)):
+            groups[name] = Group.objects.create(name=name)
+        alice = self.backend.populate_user('alice')
+        alice.groups = [groups['mirror2'], groups['mirror4']]
+
+        alice = self.backend.authenticate(username='alice', password='password')
+
+        self.assertEqual(
+            set(alice.groups.values_list("name", flat=True)),
+            {'mirror2', 'mirror3'}
+        )
+
+    def test_group_mirroring_blacklist_noop(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            GROUP_SEARCH=LDAPSearch('ou=mirror_groups,o=test', ldap.SCOPE_SUBTREE,
+                                    '(objectClass=groupOfNames)'),
+            GROUP_TYPE=GroupOfNamesType(),
+            MIRROR_GROUPS_EXCEPT=['mirror1', 'mirror2']
+        )
+
+        groups = {}
+        for name in ('mirror{}'.format(i) for i in range(1, 5)):
+            groups[name] = Group.objects.create(name=name)
+        alice = self.backend.populate_user('alice')
+        alice.groups = [groups['mirror1'], groups['mirror3']]
+
+        alice = self.backend.authenticate(username='alice', password='password')
+
+        self.assertEqual(
+            set(alice.groups.values_list("name", flat=True)),
+            {'mirror1', 'mirror3'}
+        )
+
     def test_authorize_external_users(self):
         self._init_settings(
             USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
@@ -1261,6 +1553,20 @@ class LDAPTest(TestCase):
             ['initialize', 'simple_bind_s']
         )
 
+    def test_permit_null_password(self):
+        self._init_settings(
+            USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
+            PERMIT_EMPTY_PASSWORD=True,
+        )
+
+        alice = self.backend.authenticate(username=u'alice', password=None)
+
+        self.assertEqual(alice, None)
+        self.assertEqual(
+            self.ldapobj.methods_called(),
+            ['initialize', 'simple_bind_s']
+        )
+
     def test_pickle(self):
         self._init_settings(
             USER_DN_TEMPLATE='uid=%(user)s,ou=people,o=test',
@@ -1282,6 +1588,14 @@ class LDAPTest(TestCase):
         self.assertTrue(self.backend.has_perm(alice, "auth.add_user"))
         self.assertTrue(self.backend.has_module_perms(alice, "auth"))
 
+    def test_search_attrlist(self):
+        search = LDAPSearch("ou=people,o=test", ldap.SCOPE_SUBTREE, '(uid=alice)', ['*', '+'])
+        search.execute(self.ldapobj)
+        self.assertEqual(
+            self.ldapobj.methods_called(with_args=True),
+            [('search_s', ('ou=people,o=test', ldap.SCOPE_SUBTREE, '(uid=alice)', ['*', '+']), {})]
+        )
+
     #
     # Utilities
     #
